@@ -2,8 +2,11 @@
 from __future__ import annotations
 import time 
 from colorama import Fore, Style
+import hashlib 
 import fsspec
 
+logo = f"{Fore.BLUE}{Style.BRIGHT}[Foldify]{Style.RESET_ALL}"
+
 import base64
 import zipfile
 import io
@@ -87,11 +90,12 @@ def test(seq, affinity=False):
     h = hashlib.sha256(seq.encode()).hexdigest()
     f = cache_dir / h
     if f.exists(): 
-        print(f'hit ligand test cache {seq}')
+        #print(f'hit ligand test cache {seq}')
         s = f.read_text()
         return "ok" in s, s  # cache mechanism 
     else: 
-        print(f"didn't find ligand test cache {seq}, testing...")
+        #print(f"didn't find ligand test cache {seq}, testing...")
+        pass
 
     try:
         mol = AllChem.MolFromSmiles(seq)
@@ -149,38 +153,8 @@ def fold(args):
     folder = args.input
     log = args.logmd 
 
-    if args.output == '': output_dir = f"foldify_{folder.replace('/','')}"
-    else: output_dir = args.output
-
-
-    s3_files = []
     mode = None 
     if args.s3 != '':
-        # [ ] todo remove boto3
-        import boto3
-        s3 = boto3.session.Session(profile_name="r2").client('s3', endpoint_url=args.s3)
-        resp = s3.list_objects_v2(Bucket='dmitrij', Prefix=output_dir)
-        for obj in resp.get('Contents', []): 
-            s3_files.append(obj['Key'].replace('.zip','.fasta').replace('foldify_', ''))
-
-        if args.input[:5] != "s3://": 
-            num_local = len(files)
-            num_s3 = len(s3_files)
-
-            _files = []
-            num_overlap = 0
-            for file in files: 
-                if file not in s3_files: 
-                    _files.append(file)
-                else: 
-                    num_overlap += 1
-            files = _files
-
-            print(f"Local finished: {num_local}")
-            print(f"S3 finished: {num_s3} ")
-            print(f"Overlap: {num_overlap}. ")
-            print(f"Left: {len(files)}. ")
-
         fs = fsspec.filesystem(
             "s3",
             profile="r2",
@@ -192,107 +166,72 @@ def fold(args):
                 "s3": {"addressing_style": "path"} 
             }
         )
-        print("s3")
+        print(f"{logo} Using s3 filesystem (not local). ")
         mode = S3
-
         path = args.input[5:]
     else: 
-        print("local")
+        print(f"{logo} Using local filesystem (not s3). ")
         fs = fsspec.filesystem("file")
         path = args.input
         mode = LOCAL
 
+
+    # single file 
     if args.input.endswith('.fasta'):
         files = [args.input]
         folder = folder.replace('.fasta', '')
+    # directory 
     else:
-        # [ ] port to s3 compatible. 
-        #files = [a for a in Path(folder).rglob("*.fasta")]
-        #files = sorted(files, key=lambda p: os.path.getsize(str(p)))
-        #files = [str(a) for a in files]
         files = fs.glob(f"{path}*.fasta")
         files = sorted(files, key=lambda p: int(fs.size(p)))
         path = path.split('/')
         path = path[:1] + ['foldify_' + path[1]] + path[2:]
         path = "/".join(path)
         done = fs.glob(f"{path}*.zip")
-
         zip_ids = {x.split('/')[-1].rsplit('.', 1)[0] for x in done}
 
-        if mode == S3: 
-            left = [f"s3://{f}" for f in files if f.split('/')[-1].rsplit('.', 1)[0] not in zip_ids]
-        elif mode == LOCAL:
-            left = [f for f in files if f.split('/')[-1].rsplit('.', 1)[0] not in zip_ids]
-
-        print("S3 jobs: ", len(files))
-        print("S3 done: ", len(done))
-        print("Left: ", len(left))
+        if mode == S3:      left = [f"s3://{f}" for f in files if f.split('/')[-1].rsplit('.', 1)[0] not in zip_ids]
+        elif mode == LOCAL: left = [f for f in files if f.split('/')[-1].rsplit('.', 1)[0] not in zip_ids]
 
+        print(f"{logo} Total: \t", len(files))
+        print(f"{logo} Done: \t", len(done))
+        print(f"{logo} Left: \t", len(left))
         files = left
 
-        # if s3 don't look for local files.
-        # but we want to look for stuff that's already finished. 
 
     if log: l = LogMD()
 
-
     wait_on_server(args.host)
 
+    cache_folder = f"{args.chunks}_{args.msas_num}_{args.msas_topk}_{args.msas_k}"
+    assert args.chunks == 80, "Do not support controling this from CLI yet;  MSA is pre-loaded to pinned RAM." 
+
     pbar = tqdm(files[::-1])
     for c,p in enumerate(pbar):
 
-        # 
-        if args.s3 != '': # handle single file vs folder. 
-            pass 
-            '''if "s3://" not in args.input: 
-                s3_path = path[:-1] + '.zip'
-                s3_path = '/'.join(s3_path.split('/')[0::2])
-            else:
-                s3_path = path[:-1] + '.zip'
-                s3_path = '/'.join(s3_path.split('/')[0::2])'''
+        # handle single file vs folder. 
+        if args.s3 != '': 
             s3_path = path + p.split('/')[-1].replace('.fasta', '.zip')
         else: 
             parts = p.split('/')
-            middle = ["foldify_" + parts[-2]] if args.output == '' else [args.output]
-            path = parts[:-2] + middle + [parts[-1].replace('.fasta', '')]
-            path = "/".join(path) + '/'
-
-            # add output dir to this. 
-
-        #print(path)
-        #exit()
-        #print(path, s3_path)
-        #exit()
-
-        # [ ] s3 compatible. 
-        if False: #os.path.exists(f"{path}boltz2_prediction_0.pdb") and not args.override:
-            print(f"Local: Skipping {p} found {path}boltz2_prediction_0.pdb")
-
-            # if args.s3 != '' it doesn't exist on s3 but does exist locally => upload it. 
-            # this may have killed the data i already computed. 
-            if args.s3 != '' and not file_exists(s3, 'dmitrij', s3_path):
-                print(f"Uploading {p} to s3 {s3_path}")
-                zip_bytes = io.BytesIO()
-                with zipfile.ZipFile(zip_bytes, 'w') as zip_ref:
-                    for root, dirs, files in os.walk(path):
-                        for file in files:
-                            zip_ref.write(os.path.join(root, file), os.path.relpath(os.path.join(root, file), path))
-                zip_bytes.seek(0)
-                with open('my_folder.zip', 'wb') as f: 
-                   f.write(zip_bytes.getvalue())
-                zip_bytes.seek(0)
-                s3.upload_fileobj(zip_bytes, 'dmitrij', s3_path)
-            continue 
+            if len(parts)>1:
+                middle = ["foldify_" + parts[-2]] if args.output == '' else [args.output]
+                path = parts[:-2] + middle + [parts[-1].replace('.fasta', '')]
+                path = "/".join(path) + '/'
+            else: 
+                path = p.replace('.fasta', '')
 
         try: 
             skip = False 
 
+
             p = str(p)
             #content = open(p).read()
             content = fs.open(p, "rt").read()
 
             num_tokens = sum([len(line) for line in content.split('\n') if not line.startswith('>')])
 
+            proteins = []
             for line in content.split('\n'):
                 if line.startswith('>'): continue 
                 if line == '': continue 
@@ -301,13 +240,44 @@ def fold(args):
                         print(f"Skipping {p}. RDKit didn't like {line}. ")
                         #open(new_path, 'w').write(f"Skipping {p}. RDKit didn't like {line}. ")
                         skip = True 
-                    else: print('ok')
+                    else: 
+                        #print('ok')
+                        pass
+                if get_type(line) == 'protein': proteins.append(line)
             if skip: continue 
             encoded = urllib.parse.quote(content, safe="")
             if len(content) == 0: 
                 continue 
 
-            url = f"{args.host}/fold?ui=False&only_return_zip=True&seq={encoded}&{args.args}&gpu={args.gpu}&get_msa_from_server={args.msa}"
+            if args.msa_s3 != '':
+                os.makedirs(f"sseqs/sseqs/cache_msa/{cache_folder}/", exist_ok=True)
+
+                for protein in proteins:
+                    protein_hash = hashlib.sha256(protein.encode('utf-8')).hexdigest()
+                    pth = f"sseqs/sseqs/cache_msa/{cache_folder}/{protein_hash}.a3m"
+
+                    src_path = f"{args.input[5:].split('/')[0]}/{cache_folder}/{protein_hash}.a3m"
+                    with fs.open(src_path, 'r') as src, open(pth, 'w') as dst:
+                        dst.write(src.read())
+
+                #continue 
+                # [ ] could have a preload or asynch load of all;  probably adds ~1s or so to load.
+
+            # [ ] generalize this; the arguments s
+            if args.msa_s3_precompute != '':
+                need_to_compute_msa = False
+                for protein in proteins: 
+                    protein_hash = hashlib.sha256(protein.encode('utf-8')).hexdigest()
+                    if not fs.exists(f"{args.input[5:].split('/')[0]}/{cache_folder}/{protein_hash}.a3m"): 
+                        need_to_compute_msa = True
+
+                if not need_to_compute_msa: continue 
+
+            # [ ] refactor so args dict is cast to URL. 
+            url = f"{args.host}/fold?ui=False&seq={encoded}&gpu={args.gpu}&msas_num={args.msas_num}"
+            url += f"&msas_topk={args.msas_topk}&msas_k={args.msas_k}"
+            if args.msa_s3_precompute: url += "&return_only_msa=True"
+            else: url += f"&get_msa_from_server={args.msa}&only_return_zip=True"
 
             # Open connection with progress reporting
             response = urllib.request.urlopen(url)
@@ -319,46 +289,47 @@ def fold(args):
                 if not data:
                     break
                 result += data.decode('utf-8')
-                pbar.set_description(f"{Fore.BLUE}{Style.BRIGHT}[Foldify]{Style.RESET_ALL} {time.strftime('%d-%m %H:%M:%S')} {p} tokens={num_tokens} {len(result)/1000}KB")
+                pbar.set_description(f"{logo} {time.strftime('%H:%M:%S')} {p} tokens={num_tokens} {len(result)/1000}KB")
 
             jsons = [json.loads(a) for a in result.split('\n@\n') if a != '']
-            # last is a zip file, unzip. 
-            b64_zip_data = jsons[-1]['data']
 
-            zip_bytes = base64.b64decode(b64_zip_data)
-            zip_in_memory = io.BytesIO(zip_bytes)
-            with zipfile.ZipFile(zip_in_memory, 'r') as zip_ref:
-                # [ ] remove s3 from this path 
-                # this path didn't change in each iteration? 
-                if mode == LOCAL: 
-                    os.makedirs(path, exist_ok=True)
-                    zip_ref.extractall(path) 
+            if args.msa_s3_precompute: 
+                # write the MSA cache one place?
+                for j in jsons: 
+                    if j['type'] == 'data_msa':
+                        for k,v in j['data'].items():
+                            # [ ] refactor hash / folder location to be from foldify and use that in backend
+                            os.makedirs(f"sseqs/sseqs/cache_msa/{cache_folder}/", exist_ok=True)
+                            content = v
+                            protein = v.split(">original_query\n")[1].split("\n")[0]
+                            protein_hash = hashlib.sha256(protein.encode('utf-8')).hexdigest()
+                            pth = f"sseqs/sseqs/cache_msa/{cache_folder}/{protein_hash}.a3m"
+                            open(pth, 'w'). write(content) 
+                            with fs.open(f"{args.input[5:].split('/')[0]}/{cache_folder}/{protein_hash}.a3m", 'w') as f:
+                                f.write(content)
 
-                #if args.s3 != '' and args.input[:5] != "s3://":
-                elif mode == S3: 
-                    with fs.open(f"{s3_path}", 'wb') as f:
-                        f.write(zip_bytes)
-                #    # integrate s3 with logmd -- parallel download s3 zips and store localcache? 
-                '''elif args.s3 != '':
 
-                    #import boto3
-                    # https://8a6ab2cee54f34a71f5a8d99e92da2d2.r2.cloudflarestorage.com
-                    #s3 = boto3.session.Session(profile_name="r2").client('s3', endpoint_url=args.s3)
-                    #s3.upload_file('foldify_folders.zip', 'dmitrij', 'foldify_folders.zip')
-                    zip_in_memory.seek(0)
-                    s3.upload_fileobj(zip_in_memory, 'dmitrij', s3_path) # 'foldify_folders.zip')
-                    print(s3_path)'''
+            else: 
+                b64_zip_data = jsons[-1]['data']
 
-            #time.sleep(1)
+                zip_bytes = base64.b64decode(b64_zip_data)
+                zip_in_memory = io.BytesIO(zip_bytes)
+                with zipfile.ZipFile(zip_in_memory, 'r') as zip_ref:
+                    if mode == LOCAL: 
+                        os.makedirs(path, exist_ok=True)
+                        zip_ref.extractall(path) 
+
+                    elif mode == S3: 
+                        with fs.open(f"{s3_path}", 'wb') as f:
+                            f.write(zip_bytes)
 
         except Exception as e: 
             print('something wrong', e)
             print(url)
             pass 
-        #time.sleep(1)
+
         print('')
         wait_on_server(args.host)
-        #exit()
 
 
 def main() -> None:  # pragma: no cover
@@ -378,6 +349,20 @@ def main() -> None:  # pragma: no cover
     parser.add_argument('-override', action='store_true', help='Override existing files')
     parser.add_argument('-msa', type=str, default='', help='Get msa from other ip. ')
 
+    parser.add_argument('-msa_s3_precompute', action='store_true', help='Precompute MSA and store to s3. ')
+    parser.add_argument('-msa_s3', type=str, default='', help='Load MSA from s3. ') # have one process pre-download, don't want to wait.
+
+    #foldify -i s3://dmitrij/demo.fasta -msa_s3 80_1024_20_1 -s3 https://8a6ab2cee54f34a71f5a8d99e92da2d2.r2.cloudflarestorage.com
+
+    parser.add_argument('-diffusion_samples', type=int, default=1, help='Number of boltz2 diffusion samples')
+    parser.add_argument('-msas_num', type=int, default=1024, help='Number of rows in final MSA .a3m file.')
+    parser.add_argument('-msas_k', type=int, default=1, help='Accuracy of MSA, larger is better. ')
+    parser.add_argument('-msas_topk', type=int, default=20, help='Do full smith-waterman on top_k (e.g. 20 => top 5%).')
+    parser.add_argument('-chunks', type=int, default=80, help='Cannot change this. ')
+
+
+    #-s3 https://8a6ab2cee54f34a71f5a8d99e92da2d2.r2.cloudflarestorage.com
+
     args = parser.parse_args()
 
     if args.host == 'https://gpu1.foldify.org' and not args.y:
@@ -390,6 +375,10 @@ def main() -> None:  # pragma: no cover
             print("Using remote host. ")
             print("You can skip this check with `foldify -y`")
 
+    #if args.prepare_msa and args.s3 == '':
+    #    print("Please pass in `-s3 https;//...endpoint..`, can't prepare MSAs on S3 without an endpoint. ")
+    #    exit()
+
     fold(args)
 
 
